{**
This unit realise a GUI system based on widgets.
It is completely independent from the rendering system: it doesn't know anything about how to draw itself. 
The GuiDrawer is in a separate module.

How it works:
All the classes are derived from TUiWidget, that contains some basic functionality avaiable for all widgets,
in particular all input events (clicks, keys, etc).

The TUiDesktop is a kind of manager for the gui. Applications send gathered input events to the Desktop, which 
will in turn send them to the appropriate child. 
If the input is a mouse action, desktop will send it to the child where the mouse is over (which can in turn send 
it to it's child if it's a container). If the input is a keypress, it will be sent to the currently focused widget.


Changes (17/03/05) - KAS:
  [+] TUiLayered widget added.
  [+] TUiPage widget added.
  [+] Name, Visible and Enabled properties added to TUiWidget.
  [+] Updated all constructors.
  [+] Created a WidgetByName function in TUiContainer.

  (A Visibility of false will hide it's children aswell)
  (An Enabled of false will disable it's children aswell)

  How TUiLayered works:
    The TUiLayered is derived from TUiContainer. It should ONLY have TUiPage's as it's children. Those pages will
    contain other widgets (Buttons, edits, etc). Once a TUiLayered has been populated with Pages use:

        TUiLayered(Desktop.WidgetByName('< LayeredName >')).PIndexName := '< PageName >';

    to switch to a specific page.
    A TUiLayered does not get drawn, only it's (Index) Page and children are drawn.
    Set TUiPage.DrawBack to false, if you don't want the Page to be drawn, but only the page's children will get drawn.


  Future Plans:
    [?] TUiMemo        (Multi line TUiEdit, with scroll)                        *Needed
    [?] TUiBitButton   (A button with a glyph)                                  *Handy to have
    [?] TUiMaskEdit    (For password masking, or mabye a property in TUiEdit?)
    [?] TUiScrollBox   (A container with scroll bars)
    [?] TUiImage       (??Mabye?? for easy drawing of static images?)
    [?] TUiComboBox                                                             *Handy to have
    [?] TUiCheckBox                                                             *Needed
    [?] TUiRadioButton                                                          *Needed
    
    [?] A Centered text property for most widgets (centered caption for Buttons,Labels,Edits, etc)

  Have Fun!
}
{------------------------------------------------------------------------------}
unit guiunit;

{
This unit is released under the terms of the

   GNU GENERAL PUBLIC LICENCE

as avaiable here:
http://www.gnu.org/licenses/gpl.txt

}

interface

uses Classes, SysUtils, contnrs, Nicola_xmlparser;

{** Accepted characters on editable fields }
const Characters='`1234567890-=\qwertyuiop[]asdfghjkl;''zxcvbnm,./~!@#$%^&*()_+|QWERTYUIOP{}ASDFGHJKL:"ZXCVBNM<>? ';


type

 {** Mouse button type }
 TMouseButton=(mbNone, mbLeft, mbMiddle, mbRight);

 // forward declaration
 TUiWidget = class;
 TUiContainer = class;
 TUiWindow = class;
 TUiDesktop = class;



 TOnMouseButton = procedure (Sender:TUiWidget; x,y:integer; button:TMouseButton) of object;
 TOnMouseMove = procedure (Sender:TUiWidget; x,y:integer) of object;
 {** Generic Event with no parameter but sender }
 TOnSimpleEvent = procedure (Sender:TUiWidget) of object;

 {** A List of TUiWidget  }
 TUiWidgetList = class(TList)
   function getWidget(index:integer):TUiWidget;
   property Widget[index:integer]:TUiWidget read getWidget; default;
 end;

 {** Basic widget class, from which derive all widgets }
 TUiWidget = class
   private
     fparent:TUiContainer;
     fWidth, fHeight, fTop, fLeft:integer;
     
     fOnMouseDown:TOnMouseButton;
     fOnMouseUp:TOnMouseButton;
     fOnMouseMove:TOnMouseMove;

     fUserData:pointer;

     fName:ShortString;
     fEnabled,fVisible:Boolean;
     
     
     {** Tells if the point is inside the given widget }
     function isInside(px,py:integer):boolean;
     {** Get the desktop where this widget is contained. Can return nil for widget not attached to any desktop or parent }
     function getDesktop:TUiDesktop;
     {** Tells if the widget is currently focused }
     function isFocused:boolean;

     function getWidth:integer;virtual;
     function getHeight:integer;virtual;
     procedure setWidth(v:integer);virtual;
     procedure setHeight(v:integer); virtual;

   protected
     function LoadFromNode(xml:TXmlNode):boolean; virtual;
   public

     constructor Create;virtual;

     {** Destroy the widgets. This will remove the widget from the parent too.
			Also, if this is the focused widget, the focus will be set to nil (no widget).
			The same for Desktop.WidgetOver (the widget which has the mouse on).}
     destructor Destroy;override;
     
     {** Force the focus to this widget }
     procedure Focus;
     
     {** This will fire the OnMouseDown event }
     procedure DoMouseDown(x,y:integer; button:TMouseButton); virtual;
     {** This will fire the OnMouseUp event }
     procedure DoMouseUp(x,y:integer; button:TMouseButton); virtual;
     {** This will fire the OnMouseMove event }
     procedure DoMouseMove(x,y:integer); virtual;
     {** This will fire the onKeyPress event }
     procedure DoKeyPress(k:char);virtual;
     
     {** Fire the OnMouseEnter event (todo) }
     procedure DoMouseEnter;virtual;
     {** Fire the OnMouseExit event (todo) }
     procedure DoMouseExit;virtual;
     
     {** Get the parent of the current widget }
     property Parent:TUiContainer read fParent;
     {** Get the desktop where this widget (or its containers) is. }
     property Desktop:TUiDesktop read getDesktop;
     {** Tells if the widget is currently focused }
     property Focused:boolean read isFocused;
     
     property Left:integer read fLeft write fLeft;
     property Top:integer read fTop write fTop;
     property Width:integer read getWidth write setWidth;
     property Height:integer read getHeight write setHeight;

     {** A pointer for user usage }
     property UserData:pointer read fUserData write fUserData;

     property Name:ShortString read fName write fName; //Disable write to Name??
     property Enabled:boolean read fEnabled write fEnabled;
     property Visible:boolean read fVisible write fVisible;

 end;
 
 {** A container widget, that is a widget that can contain other widgets.
  The Margin* properties define the children client area. }
 TUiContainer = class(TUiWidget)
   private
     fMarginTop,
//     fMarginBottom,
//     fMarginRight,
     fMarginLeft:integer;
     
     fChildren:TUiWidgetList;
     {** Return the widget in the given position. The coordinate are relative to this container.
     If no child is found, then self is returned (becouse the coordinates are on this container)
     If deep is specified, then is the child is in turn a container, it will search for it's own child,
     and so on.}
     function WidgetByPos(x,y:integer; deep:boolean):TUiWidget; virtual;
     function get(i:integer):TUiWidget;

   protected
     function LoadFromNode(xml:TXmlNode):boolean; override;
   public
     {** Return the container children count }
     function Count:integer;
     constructor Create;override;
     {** Destroy a container and all the children }
     destructor Destroy;override;

     {** This will call the DOMouseDown for the appropriate child.
     If the mouse is not over a child, then the signal is assumed to be for this widget }
     procedure DoMouseDown(x,y:integer; button:TMouseButton); override;
     {** This will call the DoMouseUp for the appropriate child.
     If the mouse is not over a child, then the signal is assumed to be for this widget }
     procedure DoMouseUp(x,y:integer; button:TMouseButton); override;
     {** This will call the DoMouseMove for the appropriate child.
     If the mouse is not over a child, then the signal is assumed to be for this widget }
     procedure DoMouseMove(x,y:integer); override;
     {** Add a widget. It must have position relative to this container. This set it's parent to self }
     function Add(w:TUiWidget):integer; virtual;
     {** Remove a widget from the list, setting it's parent to nil }
     function Remove(w:TUiWidget):integer; virtual;

     
     //property MarginRight:integer read fMarginRight write fMarginRight;
     property MarginLeft:integer read fMarginLeft write fMarginLeft;
     property MarginTop:integer read fMarginTop write fMarginTop;
//     property MarginBottom:integer read fMarginBottom write fMarginBottom;

     property Children[index:integer]:TUiWidget read get;default;

     {** Find a widget by Name (to change properties,etc) - KAS}
     function WidgetByName(Name:ShortString):TUiWidget;virtual;
 end;

{** This represent the "desktop": a container for windows (and other widgets).
 it also act like a "manager" for the gui. In receive all events from the main program
 and send them to children.}
 TUiDesktop = class(TUiContainer)
   private
    fOver:TUiWidget;
    fMovingWindow:TUiWindow;
    fAncX, fAncY:integer;
    fFocused:TUiWidget;
    fCurrentModal:integer;
    {** Set the focus to the passed widget, eventually firing OnFocus on it and OnFocusLost
    on the previously focused widget, if any}
    procedure setFocus(w:TUiWidget);
   public
     {** Reset the focus so that no widget is focused }
     procedure ResetFocus;
     {** Process the KeyPress event, sending it to the focused widget.
     If no widget is focused, the signal is losed }
     procedure DoKeyPress(k:char);override;
     {** Process the mouse down event. This will firstly set the focus to the clicked widget, and send it
     the event. If the click is on a window top margin, the window dragging begin and no signal is sent.
     This will also bring to front the clicked window.}
     procedure DoMouseDown(x,y:integer;button:TMouseButton);override;
     {** If dragging a window, the drag stop, else the signal is transmitted to children }
     procedure DoMouseUp(x,y:integer;button:TMouseButton);override;
     {** Act much like DoMouseDown. It also handle the WidgetOver and the mouseEnter and mouseExit events.
     If a window is currently being dragged, the signal is not transmitted to child and the window is moved.}
     procedure DoMouseMove(x,y:integer);override;
     {** Create a new desktop }
     constructor Create;override;
     
     function Add(w:TUiWidget):integer;override;
     function AddModal(w:TUiWindow):integer;
     function Remove(w:TUiWidget):integer;override;

		 function IsOverWidget:boolean;	     
 
     property WidgetOver:TUiWidget read fOver;
 end;


 {** A window. This can only have TUiDesktop as parent. }
 TUiAbstractWindow = class(TUiContainer)
   private
     modalValue:integer;
 	 public
 	   {** Tells if a point in the window is valid for start dragging it.
 	   For example, it will return true when you click on the titlebar for normal window, 
 	   but windows can eventually behave differently }
 		 function isDraggablePoint(x,y:integer):boolean; virtual; abstract;
    constructor Create;override;
 end;

 {** A window. This can only have TUiDesktop as parent. 
   The MarginTop property define the area where the window 
   can be dragged and moved }
 TUiWindow = class(TUiAbstractWindow)
   private
    ftitle:string;
   protected
     function LoadFromNode(xml:TXmlNode):boolean; override;
   public
    constructor Create;override;
    function isDraggablePoint(x,y:integer):boolean; override;
    property Title:string read fTitle write fTitle;
 end;
 
 {** A button that can be clicked by user }
 TUiButton = class(TUiWidget)
   private
     fCaption:string;
     fOnClick:TOnSimpleEvent;
   protected
     function LoadFromNode(xml:TXmlNode):boolean; override;
   public
     clicked:boolean;
     constructor Create;override;
     {** Begins the clicking of the button, setting the state to "clicked".
     This will not fire the event yet, it will be fired on the mouse up event }
     procedure DoMouseDown(x,y:integer;button:TMouseButton);override;
     {** If the state of the button is "clicked" this fires the OnClick event }
     procedure DoMouseUp(x,y:integer;button:TMouseButton);override;
     {** If you exit the button while it is clicked, the button is released without firing the OnClick.
     Else, it will perform normal OnMouseExit event }
     procedure DoMouseExit; override;
     property Caption:string read fCaption write fCaption;
     property OnClick:TOnSimpleEvent read fOnClick write fOnClick;
 end;
 
 {** A single-line editable widget that can be used to enter text }
 TUiEdit = class(TUiWidget)
   private
     fText:string;
     fOnEnter:TOnSimpleEvent;
     fMask:Boolean;
   protected
     function LoadFromNode(xml:TXmlNode):boolean; override;
   public
     constructor Create;override;
     {** Handle the key pressing, changing the displayed Text accordingly }
     procedure DoKeyPress(k:char);override;
     property Text:string read fText write fText;
     property OnEnter:TOnSimpleEvent read fOnEnter write fOnEnter;
     property Mask:Boolean read fMask write fMask;
 end;

 {** A single-line label widget }
 TUiLabel = class(TUiWidget)
   private
     fCaption:string;
   protected
     function LoadFromNode(xml:TXmlNode):boolean; override;
   public
     constructor Create;override;
     property Caption:string read fCaption write fCaption;
 end;

 TUiImage = class(TUiWidget)
   private
     fImage:LongInt;
   public
     constructor Create;override;
     property Image:longint read fImage write fImage;
 end;

TUiList = class(TUiWidget)
  private
    fStrings:TStringList;
    clicked:boolean;
    fOnClick:TOnSimpleEvent;
    fIndex:integer;
    fMarginTop:integer;
    function GetText:string;
  protected
    function LoadFromNode(xml:TXmlNode):boolean; override;
  public
    {** Behave like a simple widget that lists strings, no selecting, etc }
    Simple:boolean;
    constructor Create; override;
    {** This avoids that the widget remain clicked when the mouse exits from it. 
    Needed for widgets that act on release of mouse, like TUiButton }
    procedure DoMouseExit; override;
    procedure DoMouseDown(x,y:integer;button:TMouseButton);override;
    procedure DoMouseUp(x,y:integer; button:TMouseButton);override;
    property SelIndex:Integer read fIndex write fIndex;
    property SelText:String read GetText;
    property Strings:TStringList read fStrings;
    property MarginTop:integer read fMarginTop write fMarginTop;
    property OnClick:TOnSimpleEvent read fOnClick write fOnClick;
end;



{------------------------------------------------------------------------------}

  TUiPage = class; // forward declaration

  {** A Notebook type widget - KAS }
  TUiLayered = class (TUiContainer)
    private
      fPageIndexName:shortstring;
      procedure Switchpage(Page:shortstring);
      function FindPageIndex(Name:ShortString):integer;//find the index of a specific page
   protected
     function LoadFromNode(xml:TXmlNode):boolean; override;
    public
      constructor Create;override;
      property PIndexName: shortstring read fPageIndexName write Switchpage;//PageIndex;
  end;

  {** A page component used for the TUiLayered widget - KAS}  { simply a Container for widgets }
  TUiPage = class(TUiContainer)
    private
      fDrawBack:boolean;//use this to hide the page from being drawn, but it's children will still be drawn.
   protected
     function LoadFromNode(xml:TXmlNode):boolean; override;
   public
      constructor Create;override;
      property DrawBack:Boolean read fDrawBack write fDrawBack;
  end;
  

  {** A CheckBox widget - KAS }
  TUiCheckBox = class(TUiWidget)
    private
      fCaption:string;
      fOnToggle:TOnSimpleEvent;
      fChecked:boolean;
    protected
      function LoadFromNode(xml:TXmlNode):boolean; override;
    public
      clicked:boolean;
      constructor Create;override;
      procedure DoMouseDown(x,y:integer;button:TMouseButton);override;
      procedure DoMouseUp(x,y:integer;button:TMouseButton);override;//toggle checked
      procedure DoMouseExit; override;
      property Caption:string read fCaption write fCaption;
      property OnToggle:TOnSimpleEvent read fOnToggle write fOnToggle;
      property Checked:boolean read fChecked write fChecked;//Return check state
      {(quick xml)     <CAPTION> </CAPTION>  <CHECKED> </CHECKED>              }
  end;


  {** A ProgressBar widget - KAS }
  {Might need to add checking to see if position goes outside of min and max}
  TUiBar = class(TUiWidget)
    private
      fCaption:string;
      fShowCaption:boolean;        //Show the caption?
      fShowPosition:boolean;       //Show the position?
      fMin,fMax,fPosition:integer;
   protected
      function LoadFromNode(xml:TXmlNode):boolean; override;
    public
      constructor Create;override;
      property Caption:string read fCaption write fCaption;
      property ShowCaption:boolean read fShowCaption write fShowCaption;
      property ShowPosition:boolean read fShowPosition write fShowPosition;
      property Position:integer read fPosition write fPosition;//Current Position
      property Min:integer read fMin write fMin;               //Max value
      property Max:integer read fMax write fMax;               //Min Value
      {(quick xml)  <CAPTION> </CAPTION>  <SHOWCAPTION> </SHOWCAPTION>  <MIN> </MIN>  <MAX> </MAX> <POSITION>  </POSITION>}
   end;


  

{------------------------------------------------------------------------------}

 {** A tracker that can be dragged up and down }
 TUiTracker = class(TUiWidget)
   private
     {** What's written on the track }
     fCaption:string;
     {** if it's being moved now }
     fdragged:boolean;
     {** The position, as a value between min and max }
     fPosition:integer;
     {** define the range of values }
     fMin, fMax:integer;
     {** The width of the cursor, in graphical unit }
     fCursorWidth:integer;
     {** The margine on left and right for the cursor movement }
     fMargin:integer;
     {** event called when the position is changed }
     fOnChange:TOnSimpleEvent;
     {** Translate logic position (min>position>max) to graphic position (margin>cursor pos>width-margin) }
     function TranslateLogicToReal(v:integer):integer;
     {** Translate graphic position (margin>cursor pos>width-margin) to logic position (min>position>max) }
     function TranslateRealToLogic(v:integer):integer;
     {** Tells if the given X graphic coordinate is over the cursor }
     function IsCoordsOnCursor(x:integer):boolean;
   protected
     function LoadFromNode(xml:TXmlNode):boolean; override;
   public
     constructor Create;override;
     
     procedure DoMouseMove(x,y:integer);override;
     procedure DoMouseDown(x,y:integer;button:TMouseButton);override;
     procedure DoMouseUp(x,y:integer;button:TMouseButton);override;
     procedure DoMouseExit; override;
     
     property Caption:string read fCaption write fCaption;
     property CursorWidth:integer read fCursorWidth write fCursorWidth;
     property Margin:integer read fMargin write fMargin;
     property Min:integer read fMin write fMin;
     property Max:integer read fMax write fMax;
     property Position:integer read fPosition write fPosition;
     property OnChange:TOnSimpleEvent read fOnChange write fOnChange;
 end;


 TUiRollableWindow = class(TUiWindow)
   private
    fOpen:boolean;
    function getHeight:integer; override;
   protected
     function LoadFromNode(xml:TXmlNode):boolean; override;
   public
    constructor Create;override;
    procedure DoMouseDown(x,y:integer;button:TMouseButton);override;
    property Open:boolean read fOpen write fOpen;
 end;

{** A cursor for the multi tracker }
TCursor = class
  private
    fPosition:integer;
    fData:TObject;
  public
    property Position:integer read fPosition write fPosition;
    property Data:TObject read fData write fData;
end;


{** A cursor list for the multi tracker }
TCursorList = class(TObjectList)
  private
    function getCursor(index:integer):TCursor;
  public
    property Cursor[index:integer]:TCursor read getCursor; default;
end;

TUiMultiTracker = class(TUiWidget)
  private
     fcursors:TCursorList;
     {** what tracker is being moved }
     fdragged:integer;
     {** define the range of values }
     fMin, fMax:integer;
     {** The width of the cursor, in graphical unit }
     fCursorWidth:integer;
     {** The margine on left and right for the cursor movement }
     fMargin:integer;
     {** event called when the position is changed }
     fOnChange:TOnSimpleEvent;
     {** event called when the a different cursor is selected }
     fOnSelect:TOnSimpleEvent;
     {** index of selected cursor }
     fSelIndex:integer;
     {** Translate logic position (min>position>max) to graphic position (margin>cursor pos>width-margin) }
     function TranslateLogicToReal(v:integer):integer;
     {** Translate graphic position (margin>cursor pos>width-margin) to logic position (min>position>max) }
     function TranslateRealToLogic(v:integer):integer;
     {** Tells if the given X graphic coordinate is over a cursor }
     function IsCoordsOnCursor(x:integer):integer;
  public
    constructor Create;
    destructor Destroy;override;
    function AddCursor:TCursor;

    procedure DoMouseMove(x,y:integer);override;
    procedure DoMouseDown(x,y:integer;button:TMouseButton);override;
    procedure DoMouseUp(x,y:integer;button:TMouseButton);override;
    procedure DoMouseExit; override;

    function LoadFromNode(xml:TXmlNode):boolean; override;

    property CursorWidth:integer read fCursorWidth write fCursorWidth;
    property Margin:integer read fMargin write fMargin;
    property Min:integer read fMin write fMin;
    property Max:integer read fMax write fMax;
    property Cursors:TCursorList read fcursors;
    property SelIndex:integer read fSelIndex write fSelIndex;

    property OnChange:TOnSimpleEvent read fOnChange write fOnChange;
    property OnSelect:TOnSimpleEvent read fOnSelect write fOnSelect;

end;





function LoadGui(parent:TUiContainer; filename:TFilename):boolean; overload;
function LoadGui(parent:TUiContainer; node:TXmlNode):boolean;overload;
function LoadWidget(filename:TFilename; WidgetName:string):TUiWidget;overload;
function LoadWidget(node:TXmlNode; WidgetName:string):TUiWidget;overload;

implementation

function CreateWidget(n:TXmlNode):TUiWidget;
begin
if n.name='WINDOW' then result:=TUiWindow.create
else if n.name='ROLLABLEWINDOW'  then result:=TUiRollableWindow.create
else if n.name='BUTTON'  then result:=TUiButton.create 
else if n.name='LAYERED' then result:=TUiLayered.create
else if n.name='PAGE' then result:=TUiPage.create
else if n.name='TRACKER' then result:=TUiTracker.create
else if n.name='EDIT' then result:=TUiEdit.create
else if n.name='LIST' then result:=TUiList.create
else if n.name='IMAGE' then result:=TUiImage.create
else if n.name='LABEL' then result:=TUiLabel.create
else if n.name='CHECKBOX' then result:=TUiCheckBox.create
else if n.name='BAR' then result:=TUiBar.create
else if n.name='MULTITRACKER' then result:=TUiMultiTracker.create

else result:=nil;
end;


function LoadWidget(node:TXmlNode; WidgetName:string):TUiWidget;overload;
var i:integer;
    n:TXmlNode;
		w:TUiWidget;
begin
  result:=nil;
  for i:=0 to node.children.count-1 do
    begin
    	n:=node.children[i];
    	if n.nodetype=ndTag then
       if n.Attributes['name']=WidgetName then
    	  begin
          // if this is the required widget, laod & return it.
    	    result:=CreateWidget(n);
    	    if result<>nil then result.loadFromNode(n);
          exit;
    	  end;
    end;
end;

function LoadGui(parent:TUiContainer; node:TXmlNode):boolean;
var i:integer;
    n:TXmlNode;
		w:TUiWidget;
begin
  for i:=0 to node.children.count-1 do
    begin
    	n:=node.children[i];
    	if n.nodetype=ndTag then
    	  begin
    	    w:=CreateWidget(n);
    	    if w<>nil then result:=w.loadFromNode(n)
    	              else result:=false;
    	              
    	    if result then parent.add(w)
    	    					else exit;
    	  end;
    end;
end;

function LoadGui(parent:TUiContainer; filename:TFilename):boolean;
var d:TXmlDocument;
begin
	result:=false;
	try
	  d:=TXmlDocument.create;
	  d.load(filename);
	  result:=LoadGui(parent, d.node);
	finally
	  d.free;
  end;
end;

function LoadWidget(filename:TFilename; WidgetName:string):TUiWidget;
var d:TXmlDocument;
begin
	result:=nil;
	try
	  d:=TXmlDocument.create;
	  d.load(filename);
	  result:=LoadWidget(d.node, WidgetName);
	finally
	  d.free;
  end;
end;





{** Process a keypress on a string buffer. This means adding printable characters and handling correctly other
keys such as delete, backspace, etc, etc }
procedure HandleKeyPress(var buffer:string; key:char);
begin
if Pos(Key,Characters)>0 then buffer:=buffer+key
else
begin
 case key of
   #8: // backspace
    if length(buffer)> 0 then buffer:=copy(buffer, 1, length(buffer)-1);
 end;
end;
end;


function TUiWidgetList.getWidget(index:integer):TUiWidget;
begin
 result:=TUiWidget(items[index]);
end;


constructor TUiWidget.Create;
begin
inherited Create;
width:=100;
height:=50;
left:=30;
top:=0;

Enabled:=true;
visible:=true;
end;

constructor TUiButton.Create;
begin
inherited Create;
fName:='Button1';
fCaption:=fName;
width:=100;
height:=50;
left:=0;
top:=0;
end;

constructor TUiEdit.Create;
begin
inherited Create;
fName:='Edit1';
ftext:='';
width:=100;
height:=50;
left:=0;
top:=0;
end;

constructor TUiLabel.Create;
begin
inherited Create;
fName:='Label1';
fCaption:='';
width:=100;
height:=50;
left:=0;
top:=0;
end;

constructor TUiImage.Create;
begin
inherited Create;
fName:='Image1';
fImage:=0;
width:=100;
height:=50;
left:=0;
top:=0;
end;


destructor TUiWidget.Destroy;
var d:TUiDesktop;
begin
d:=getDesktop;
if d<>nil then
 begin
   if d.fOver=self then d.fOver:=nil;
   if d.fFocused=self then d.setFocus(nil);
 end;
if Parent<>nil then
  Parent.Remove(self);

inherited Destroy;
end;

constructor TUiContainer.Create;
begin
inherited Create;
MarginLeft:=0;
MarginTop:=0;
fChildren:=TUiWidgetList.create;
end;

destructor TUiContainer.Destroy;
var i:integer;
    w:tuiwidget;
begin
  while fChildren.count>0 do
    begin
      fChildren[0].free;
    end;
  inherited Destroy;
end;

function TUiContainer.Remove(w:TUiWidget):integer;
begin
result:=fChildren.Remove(w);
w.fparent:=nil;
end;


function TUiContainer.Add(w:TUiWidget):integer;
begin
result:=fChildren.Add(w);
w.fParent:=self;
end;

constructor TUiWindow.Create;
begin
inherited Create;
fName:='Window1';
end;

constructor TUiAbstractWindow.Create;
begin
inherited Create;
ModalValue:=0;
end;

function TUiWidget.isInside(px,py:integer):boolean;
begin
 result := (px>=left) and (px<=left+Width) and
           (py>=top) and (py<=top+Height);
end;

procedure TUiWidget.DoKeyPress(k:char);
begin
  // @todo call onKeyPress event
end;

procedure TUiWidget.DoMouseDown(x,y:integer; button:TMouseButton);
begin
  if assigned(fOnMouseDown) then fOnMouseDown(self, x,y,button);
end;

procedure TUiWidget.DoMouseUp(x,y:integer; button:TMouseButton);
begin
  if assigned(fOnMouseUp) then fOnMouseUp(self, x,y,button);
end;

procedure TUiWidget.DoMouseMove(x,y:integer);
begin
  if assigned(fOnMouseMove) then fOnMouseMove(self, x,y);
end;

{ Get a widget by Name - KAS }
{
  This function will search all widgets and children for a requested widget.
}
function TUiContainer.WidgetByName(Name:ShortString):TUiWidget;
var i:integer;
    w:TUiWidget;
begin
   result:=nil;

   for i:=fChildren.Count-1 downto 0 do begin                                   //search children
     w:=fChildren[i];
     if LowerCase(w.Name) = LowerCase(Name) then begin                          //if names match it's found
        result:=w;
        exit;
     end
     else begin
        if (w is TUiContainer) then                                             //if a container, then search it's children
          result:=TUiContainer(w).WidgetByName(Name);
          if result <> nil then exit;                                           //if it returns something, then the requested
     end;                                                                       //widget was found, so stop searching.
   end;

end;

procedure TUiContainer.DoMouseDown(x,y:integer; button:TMouseButton);
var w:TUiWidget;
    tx,ty:integer;
begin
tx:=x-MarginLeft;
ty:=y-MarginTop;

// find the child contained
w:=WidgetByPos(tx,ty, false);
if w=self then
 begin
  // no child found! the event was for us
  inherited DoMouseDown(x,y,button); // call onMouseDown by default :P
 end
else
 begin
  w.DoMouseDown(tx - w.Left,ty - w.Top,button); // translate even more to child local
 end;

end;

procedure TUiContainer.DoMouseUp(x,y:integer; button:TMouseButton);
var w:TUiWidget;
    tx,ty:integer;
begin
tx:=x-MarginLeft;
ty:=y-MarginTop;

// find the child contained
w:=WidgetByPos(tx,ty, false);
if w=self then
 begin
  // no child found! the event was for us
  inherited DoMouseUp(x,y,button); // call onMouseUp by default :P
 end
else
 begin
  w.DoMouseUp(tx - w.Left,ty - w.Top,button); // translate even more to child local
 end;

end;

procedure TUiContainer.DoMouseMove(x,y:integer);
var w:TUiWidget;
    tx,ty:integer;
begin
tx:=x-MarginLeft;
ty:=y-MarginTop;

// find the child contained
w:=WidgetByPos(tx,ty, false);
if w=self then
 begin
  // no child found! the event was for us
  inherited DoMouseMove(x,y); // call onMouseMove by default :P
 end
else
 begin
  w.DoMouseMove(tx - w.Left,ty - w.Top); // translate even more to child local
 end;

end;



function TUiContainer.WidgetByPos(x,y:integer; deep:boolean):TUiWidget;
var i:integer;
    w:TUiWidget;

begin
   result:=self;
   for i:=fChildren.Count-1 downto 0 do
    begin
     w:=fChildren[i];
     if (w.Enabled=false) or (w.Visible=false) then continue;                   //if Disabled or invisible then skip
     if w.isInside(x, y) then
      begin
        if deep and (w is TUiContainer) then begin
          result:=TUiContainer(w).WidgetByPos(x-TUiContainer(w).MarginLeft-TUiContainer(w).left,y-TUiContainer(w).margintop-TUiContainer(w).top,deep);
        end
       else
          result:=w;
        exit;
      end;
    end;
end;

function TUiContainer.Count:integer;
begin
result:=fChildren.count;
end;

function TUiContainer.get(i:integer):TUiWidget;
begin
result:=fChildren[i];
end;

constructor TUiDesktop.Create;
begin
inherited create;
fCurrentModal:=0;
fName:='Desktop1';
fOver:=self;
setFocus(nil);
fparent:=nil;
left:=0;
top:=0;
marginleft:=0;
margintop:=0;
end;

procedure TUiDesktop.DoKeyPress(k:char);
begin
if (fFocused = nil) or (fFocused = self) then exit;
fFocused.DoKeyPress(k);
end;

procedure TUiDesktop.DoMouseDown(x,y:integer;button:tMouseButton);
var w:TUiWidget;
    tx,ty:integer;
begin
// the widget clicked
w:=WidgetByPos(x-MarginLeft, y-MarginTop, false);

// check for modals
if w is TUiAbstractWindow then
  begin
    // if the window modalvalue is less than the current, exit
    if TUiAbstractWindow(w).modalValue<fCurrentModal then exit;
  end
else
  // if we have a current value and a widget (no window) is clicked, exit
  if fCurrentModal>0 then exit;

// focus clicked deep widget
setFocus( WidgetByPos(x-MarginLeft, y-MarginTop, true) );

// first, bring to front if it's a window
if w is TUiWindow then
 fChildren.Move(fChildren.IndexOf(w), fChildren.Count-1);


tx:=x-MarginLeft-Left-TUiWindow(w).Left;
ty:=y-MarginTop-Top-TUiWindow(w).top;
// @todo replace the margintop control with something like TUiWindow.isDraggablePoint(x,y),
// so that a window can decide where the user should drag it.
if (w is TUiWindow) and (TUiWindow(w).isDraggablePoint(tx,ty)) and (button=mbLeft) then
 begin
  if button=mbLeft then
   begin
    fMovingWindow:=TUiWindow(w);
    fAncX:=x-fMovingWindow.left;
    fAncy:=y-fMovingWindow.top;
   end;
 end
else
  inherited DoMouseDown(x,y,button);

end;



procedure TUiDesktop.DoMouseMove(x,y:integer);
var w:TUiWidget;
begin
// if dragging a window, move it
if fMovingWindow<>nil then
 begin
  fMovingWindow.left:=(x-fAncX);
  fMovingWindow.top:=(y-fAncy);
 end
else
 begin
  // find the widget we are over with the mouse
  w:=WidgetByPos(x-MarginLeft, y-MarginTop, true);

  // check for modals
  if w is TUiAbstractWindow then
    begin
      // if the window modalvalue is less than the current, exit
      if TUiAbstractWindow(w).modalValue<fCurrentModal then exit;
    end
  else
    // if we have a current value and a widget (no window) is clicked, exit
    // (widget always have a modalvalue of 0)
    if fCurrentModal>0 then exit;


  // Eventually change the WidgetOver widget
  if fover<>w then
   begin
    if assigned(fOver) then fOver.DoMouseExit;
    w.DoMouseEnter;
    fOver:=w;
   end;
   inherited DoMouseMove(x,y);
 end;
 
end;

procedure TUiDesktop.DoMouseUp(x,y:integer;button:TMouseButton);
begin
if fMovingWindow<>nil then fMovingWindow:=nil
                      else inherited DoMouseUp(x,y,button);
end;

procedure TUiButton.DoMouseDown(x,y:integer; button:TMouseButton);
begin
clicked:=true;
end;

procedure TUiButton.DoMouseUp(x,y:integer; button:TMouseButton);
begin
if clicked then
 begin
   if assigned(fOnClick) then fOnClick(self);
   clicked:=false;
 end;
end;

procedure TUiWidget.DoMouseEnter;
begin
// @todo add OnMouseEnter event
end;

procedure TUiWidget.DoMouseExit;
begin
// @todo add OnMouseExit event
end;

procedure TUiButton.DoMouseExit;
begin
if clicked then clicked:=false
else inherited DoMouseExit;
end;

function TUiWidget.getDesktop:TUiDesktop;
begin
if parent = nil then result:=nil
else
    if parent is TUiDesktop then result:=TUiDesktop(parent) else result:=parent.getDesktop;
end;


function TUiWidget.isFocused:boolean;
begin
result := getDesktop.fFocused = self;
end;

procedure TUiWidget.Focus;
begin
	getDesktop.setFocus(self);
end;

procedure TUiEdit.DoKeyPress(k:char);
begin
if k = #13 then
 begin
  if assigned(fOnEnter) then fOnEnter(self);
 end
else
   HandleKeyPress(ftext, k);
end;

procedure TUiDesktop.ResetFocus;
begin
setFocus(nil);
end;

procedure TUiDesktop.setFocus(w:TUiWidget);
begin
	fFocused:=w;
	// @todo here we could eventually send OnFocus and OnFocusLost
end;

function TUiWindow.isDraggablePoint(x,y:integer):boolean; 
begin
  result := (y<=MarginTop);
end;


{---------------------------Functions-For-TUiLayered---------------------------}
                                    { KAS }
constructor TUiLayered.Create;
begin
  inherited Create;
  fName:='Layered1';
  width:=100;
  height:=100;
  marginleft:=0;
  margintop:=0;
  left:=0;
  top:=0;
end;

constructor TUiPage.Create;
begin
  inherited Create;
  fName:='Page1';
  marginleft:=0;
  margintop:=0;
  width:=100;
  height:=100;
  left:=0;
  top:=0;
  DrawBack:=false;
end;

function TUiLayered.FindPageIndex(Name:ShortString):integer;
var i:integer;
begin
  result := 0;
  for i := 0 to fChildren.Count-1 do begin
    if LowerCase(fChildren[i].Name) = LowerCase(Name) then begin
      result := i;
      exit;
    end;
  end;
end;

procedure TUiLayered.Switchpage(Page:shortstring);
begin
  if LowerCase(Page) <> LowerCase(PIndexName) then begin 
    fPageIndexName := Page;
    FChildren.Move(FindPageIndex(LowerCase(Page)), fChildren.Count-1);//Move the requested page to the front
  end;
end;


{------------------------------------------------------------------------------}

function TUiWidget.LoadFromNode(xml:TXmlNode):boolean; 
var n:TXmlNode;
begin
name:=xml.attributes['name'];
	
n:=xml.query('LEFT');
if n<>nil then left:=n.Number;

n:=xml.query('TOP');
if n<>nil then top:=n.Number;

n:=xml.query('WIDTH');
if n<>nil then width:=n.Number;

n:=xml.query('HEIGHT');
if n<>nil then height:=n.Number;

n:=xml.query('NAME');
if n<>nil then name:=n.Text;

n:=xml.query('VISIBLE');
if n<>nil then visible:=(lowercase(n.Text)='true');

n:=xml.query('ENABLED');
if n<>nil then enabled:=(lowercase(n.Text)='true');

result:=true;
end;

function TUiContainer.LoadFromNode(xml:TXmlNode):boolean;
var n:TXmlNode;
begin
	
result:=true;	
	
n:=xml.query('MARGINLEFT');
if n<>nil then marginleft:=n.Number;

n:=xml.query('MARGINTOP');
if n<>nil then margintop:=n.Number;

n:=xml.query('CHILDREN');
if n<>nil then 
 begin
 	 result:=LoadGui(self, n);
 end;

result:=inherited LoadFromNode(xml);

end;

function TUiWindow.LoadFromNode(xml:TXmlNode):boolean; 
var n:TXmlNode;
begin
	
n:=xml.query('TITLE');
if n<>nil then title:=n.text;

result:=inherited LoadFromNode(xml);
end;

function TUiButton.LoadFromNode(xml:TXmlNode):boolean; 
var n:TXmlNode;
begin
 n:=xml.query('CAPTION');
 if n<>nil then caption:=n.text;
 
 result:=inherited LoadFromNode(xml);
end;

function TUiLabel.LoadFromNode(xml:TXmlNode):boolean; 
var n:TXmlNode;
begin
 n:=xml.query('CAPTION');
 if n<>nil then caption:=n.text;
 
 result:=inherited LoadFromNode(xml);
end;

function TUiEdit.LoadFromNode(xml:TXmlNode):boolean; 
var n:TXmlNode;
begin
 n:=xml.query('TEXT');
 if n<>nil then text:=n.text;
 
  n:=xml.query('MASK');
  if n<>nil then mask:=(lowercase(n.Text)='true');
 
 
 result:=inherited LoadFromNode(xml);
end;

function TUiList.LoadFromNode(xml:TXmlNode):boolean; 
var n:TXmlNode;
begin
 n:=xml.query('TEXT');
 if n<>nil then strings.text:=n.text;

 n:=xml.query('MARGINTOP');
 if n<>nil then fmargintop:=n.number;

  n:=xml.query('SIMPLE');
  if n<>nil then simple:=(lowercase(n.Text)='true');
 
 
 result:=inherited LoadFromNode(xml);
end;

function TUiPage.LoadFromNode(xml:TXmlNode):boolean; 
var n:TXmlNode;
begin
  n:=xml.query('DRAWBACK');
  if n<>nil then drawback:=(lowercase(n.Text)='true');
  
  result:=inherited LoadFromNode(xml);
end;

function TUiLayered.LoadFromNode(xml:TXmlNode):boolean; 
var n:TXmlNode;
begin
  result:=inherited LoadFromNode(xml);
  if result then
   begin
	  n:=xml.query('ACTIVE');
	  if n<>nil then PIndexName := n.Text;
   end;
  
  
end;

constructor TUiList.Create;
begin
inherited Create;
fStrings:=TStringList.Create;
fIndex:=-1;
fMarginTop:=12;
Simple:=false;
end;


function TUiDesktop.Add(w:TUiWidget):integer;
begin
if w is TUiAbstractWindow then
 begin
   // if adding a window, check if his modal value is greater than the current one.
   // if it is, use window's modal value as the current.
   if TUiAbstractWindow(w).modalValue>fCurrentModal then fCurrentModal:= TUiAbstractWindow(w).modalValue;
 end;
result:=inherited Add(w);
end;

function TUiDesktop.Remove(w:TUiWidget):integer;
var i,m:integer;
begin
result:=inherited Remove(w);
// if it was a window, recalculate current max modal value
if w is TUiAbstractWindow then
 begin
   m:=0;
   for i:=0 to count-1 do
    if children[i] is TUiAbstractWindow then
     if TUiAbstractWindow(children[i]).modalValue>m then
      m:=TUiAbstractWindow(children[i]).modalValue;
   fCurrentModal:=m;
 end;
end;

function TUiDesktop.AddModal(w:TUiWindow):integer;
var i:integer;
begin
 // set the modal value as the greates
 w.modalValue:=fCurrentModal+1;
 // add it
 result:=add(w);
 // remove focus (that surely was on some non-modal widget)
 setFocus(nil);
end;

procedure TUiList.DoMouseDown(x,y:integer; button:TMouseButton);
begin
  if not Simple then
    clicked:=true;
end; 

procedure TUiList.DoMouseUp(x,y:integer; button:TMouseButton);
var i : integer;
begin
  if (clicked) and (not Simple) then
  begin
    { Get the selected string }
    i:=(y-margintop) div 12;

    if (i >= Strings.Count) then i:=Strings.Count-1
       else if i<0 then i:=0;

    SelIndex:=i;

    if assigned(fOnClick) then fOnClick(self);
    clicked:=false;
  end;
end;

{** This avoids that the widget remain clicked when the mouse exits from it }
procedure TUiList.DoMouseExit;
begin
  if clicked then clicked:=false
    else inherited DoMouseExit;
end;


function TUiList.GetText:string;
begin
  result:=Strings[SelIndex];
end;

procedure TUiTracker.DoMouseDown(x,y:integer; button:TMouseButton);
begin
if isCoordsOnCursor(x) then fDragged:=true
else
 if x<(width div 2) then dec(fposition) else inc(fposition);
 if assigned(fOnChange) then fOnChange(self);
end;

function TUiTracker.IsCoordsOnCursor(x:integer):boolean;
var cursorrealpos:integer;
begin
	cursorrealpos := TranslateLogicToReal(position);
	result := abs(x-cursorrealpos)<(cursorwidth div 2);
end;

function TUiTracker.TranslateLogicToReal(v:integer):integer;
var logicPos, logicWidth, realWidth:integer	;
begin
logicPos := v - min;
logicWidth := max-min;
realwidth	:=width-2*margin;

result:=(logicPos*realwidth div logicwidth) + margin;
end;

function TUiTracker.TranslateRealToLogic(v:integer):integer;
var realMin, realMax, realwidth, logicWidth, a, b:integer;
begin
	realmin:=margin;
	realmax:=width-margin;
	realwidth:=realmax-realmin;
	
	logicWidth:=max-min;
	
	b:=v-realmin; // real position without considering min
	
	a := b * logicwidth div realwidth; // logic position in width (no min)
	
	result:=a+min;
end;

procedure TUiTracker.DoMouseMove(x,y:integer);
begin
if FDragged then
 begin
 	fPosition:=TranslateRealToLogic(x);
 	if Fposition>max then fposition:=max else if fposition<min then fposition:=min;
 	if assigned(fOnChange) then fOnChange(self);
 end;
end;

procedure TUiTracker.DoMouseUp(x,y:integer; button:TMouseButton);
begin
if fDragged then
 begin
   fdragged:=false;
 end;
end;

procedure TUiTracker.DoMouseExit;
begin
	//maybe we can find a better solution for this ?
	if fDragged then fDragged:=false
	else
	inherited DoMouseExit;
end;

constructor TUiTracker.Create;
begin
  inherited Create;
  fName:='Tracker1';
  margin:=20;
  width:=100;
  height:=100;
  left:=0;
  top:=0;
  position:=0;
  fCursorWidth:=16;
  fmin:=0;
  fmax:=10;
end;

function TUiTracker.LoadFromNode(xml:TXmlNode):boolean;
var n:TXmlNode;
begin
	
result:=true;	
	
n:=xml.query('MIN');
if n<>nil then min:=n.Number;	

n:=xml.query('MAX');
if n<>nil then MAX:=n.Number;
	
n:=xml.query('POSITION');
if n<>nil then position:=n.Number;

n:=xml.query('CURSORWIDTH');
if n<>nil then cursorwidth:=n.Number;

n:=xml.query('MARGIN');
if n<>nil then margin:=n.Number;

n:=xml.query('CAPTION');
if n<>nil then caption:=n.text;

result:=inherited LoadFromNode(xml);

end;

function TUiDesktop.IsOverWidget:boolean;	     
begin
	result:=WidgetOver <> self;
end;





{--------------------------Functions-For-TUiCheckBox---------------------------}
                                    { KAS }

constructor TUiCheckBox.Create;
begin
  inherited Create;
  fName:='CheckBox1';
  fCaption:=fName;
  fChecked:=false;
  width:=100;
  height:=50;
  left:=0;
  top:=0;
end;

procedure TUiCheckBox.DoMouseDown(x,y:integer; button:TMouseButton);
begin
  clicked:=true;
end;

procedure TUiCheckBox.DoMouseUp(x,y:integer; button:TMouseButton);
begin
if clicked then
 begin
   clicked:=false;
   fChecked:=not fChecked;//Toggle Checked;
   if assigned(fOnToggle) then fOnToggle(self);
 end;
end;

procedure TUiCheckBox.DoMouseExit;
begin
  if clicked then clicked:=false
  else inherited DoMouseExit;
end;

function TUiCheckBox.LoadFromNode(xml:TXmlNode):boolean;
var n:TXmlNode;
begin
  n:=xml.query('CAPTION');
  if n<>nil then Caption:=n.text;

  n:=xml.query('CHECKED');
  if n<>nil then Checked:=(lowercase(n.Text)='true');

  result:=inherited LoadFromNode(xml);
end;

{-----------------------------Functions-For-TUiBar-----------------------------}
                                    { KAS }

constructor TUiBar.Create;
begin
  inherited Create;
  fName:='Bar1';
  fCaption:=fName;
  fShowCaption:=false;
  fShowPosition:=true;
  fMin:=0;
  fMax:=100;
  fPosition:=0;
  width:=100;
  height:=50;
  left:=0;
  top:=0;
end;

function TUiBar.LoadFromNode(xml:TXmlNode):boolean;
var n:TXmlNode;
begin
  n:=xml.query('CAPTION');
  if n<>nil then Caption:=n.text;
  n:=xml.query('SHOWCAPTION');
  if n<>nil then ShowCaption:=(lowercase(n.Text)='true');
  n:=xml.query('SHOWPOSITION');
  if n<>nil then ShowPosition:=(lowercase(n.Text)='true');
  n:=xml.query('MIN');
  if n<>nil then Min:=n.Number;
  n:=xml.query('MAX');
  if n<>nil then Max:=n.Number;
  n:=xml.query('POSITION');
  if n<>nil then Position:=n.Number;

  result:=inherited LoadFromNode(xml);
end;


function TUiWidget.getWidth:integer;
begin
result:=fWidth;
end;
function TUiWidget.getHeight:integer;
begin
result:=fHeight;
end;
procedure TUiWidget.setWidth(v:integer);
begin
fWidth:=v;
end;
procedure TUiWidget.setHeight(v:integer);
begin
fHeight:=v;
end;


constructor TUiRollableWindow.Create;
begin
 inherited Create;
 fOpen:=false;
end;

procedure TUiRollableWindow.DoMouseDown(x,y:integer;button:TMouseButton);
begin
inherited;
if button = mbRight then
  if isDraggablePoint(x,y) then
    fOpen:=not fOpen;
end;

function TUiRollableWindow.LoadFromNode(xml:TXmlNode):boolean;
var n:TXmlNode;
begin
  n:=xml.query('OPEN');
  if n<>nil then fOpen:=(lowercase(n.Text)='true');

  result:=inherited LoadFromNode(xml);
end;

function TUiRollableWindow.getHeight:integer;
begin
  if open then result:=inherited getHeight else result:=MarginTop;
end;

constructor TUiMultiTracker.Create;
begin
  inherited Create;
  Name:='MultiTracker1';
  margin:=20;
  width:=100;
  height:=16;
  left:=0;
  top:=0;
  fCursorWidth:=16;
  fmin:=0;
  fmax:=10;
  fdragged:=-1;
  fSelIndex:=-1;
  fcursors:=TCursorList.create;
end;

destructor TUiMultiTracker.Destroy;
begin
cursors.free;
inherited;
end;

function TUiMultiTracker.IsCoordsOnCursor(x:integer):integer;
var cursorrealpos:integer;
    i:integer;
begin
 for i:=0 to cursors.count-1 do
   begin
     cursorrealpos := TranslateLogicToReal(cursors[i].position);
     if abs(x-cursorrealpos)<(cursorwidth div 2) then
       begin
         result:=i;
         exit;
       end;
   end;
 result:=-1;
end;

function TUiMultiTracker.TranslateLogicToReal(v:integer):integer;
var logicPos, logicWidth, realWidth:integer	;
begin
logicPos := v - min;
logicWidth := max-min;
realwidth	:=width-2*margin;

result:=(logicPos*realwidth div logicwidth) + margin;
end;

function TUiMultiTracker.TranslateRealToLogic(v:integer):integer;
var realMin, realMax, realwidth, logicWidth, a, b:integer;
begin
	realmin:=margin;
	realmax:=width-margin;
	realwidth:=realmax-realmin;

	logicWidth:=max-min;

	b:=v-realmin; // real position without considering min

	a := b * logicwidth div realwidth; // logic position in width (no min)

	result:=a+min;
end;


function TCursorList.getCursor(index:integer):TCursor;
begin
 result:=TCursor(items[index]);
end;

procedure TUiMultiTracker.DoMouseDown(x,y:integer; button:TMouseButton);
var c:integer;
begin
c:=isCoordsOnCursor(x);
if c>=0 then
  begin
    Cursors.Move(c, 0); // this is so that it gets drawn and selected first
    fDragged:=0;        // in this way it always drags 0 indexed cursor
    fSelIndex:=0;
    if assigned(fOnSelect) then fOnSelect(self);
  end;
end;


procedure TUiMultiTracker.DoMouseMove(x,y:integer);
var p:integer;
begin
if FDragged<>-1 then
 begin
   p:=TranslateRealToLogic(x);
   if p>max then p:=max else if p<min then p:=min;

   cursors[fDragged].Position:=p;
   if assigned(fOnChange) then fOnChange(self);
 end;
end;

procedure TUiMultiTracker.DoMouseUp(x,y:integer; button:TMouseButton);
begin
if fDragged>=0 then
 begin
   fdragged:=-1;
 end;
end;

procedure TUiMultiTracker.DoMouseExit;
begin
  //maybe we can find a better solution for this ?
  if fDragged>=0 then fDragged:=-1
  else
    inherited DoMouseExit;
end;

function TUiMultiTracker.AddCursor:TCursor;
begin
 result:=TCursor.create;
 cursors.Add(result);
end;


function TUiMultiTracker.LoadFromNode(xml:TXmlNode):boolean;
var n:TXmlNode;
begin

result:=true;

n:=xml.query('MIN');
if n<>nil then min:=n.Number;

n:=xml.query('MAX');
if n<>nil then MAX:=n.Number;

n:=xml.query('CURSORWIDTH');
if n<>nil then cursorwidth:=n.Number;

n:=xml.query('MARGIN');
if n<>nil then margin:=n.Number;

result:=inherited LoadFromNode(xml);

end;


end.
