{** This unit classes to handle retrieval and loading of media data.
You just pass the name of the resource and the manager gives you the
correct resource.
Actually resources are of type TMesh and TTexture, but models, sound 
and music could be added soon.
The media manager can obtain data from files (as it actually is for TFileMediaManager) or 
from other ways. For example you can define a manager that obtain data from
a packed file containing all game media.
You can use TFileMediaManager while developing for quick media file editing, 
and than switch with a single change to a packed file manager. }
unit mediamanagerunit;


(*

This unit was written by Nicola {MSX} Lugato.

http://msx.rules.it
nicola@lugato.net
nicola.lugato@gmail.com

It is released under the term of
The GNU General Public License as published by the Free
Software Foundation.

*)


interface

uses
  glmeshunit,
  meshunit,
  soundunit,
  textureunit;

const MEDIA_PATH = 'media/';

type

    {** this is a Generic media manager, that is
        the abstract "interface" of all the media manager }
    TGenericMediaManager = class
       function getMeshByName(name:string):TMesh; virtual; abstract;
       function getTextureByName(name:string):TTexture;  virtual; abstract;
       function getSoundByName(name:string):TSound;  virtual; abstract;
    end;

    {** this is a cached media manager, an abstract manager
       that caches the file already loaded, and gives them back
       if they're requested again.
       Child of this class should override the loadXXX method to
       read the data first time. }
    TCachedMediaManager = class(TGenericMediaManager)
      private
       MeshList:TMeshList;
       SoundList:TSoundList;
       TextureList:TTextureList;
       function loadMesh(name:string):TMesh; virtual; abstract;
       function loadTexture(name:string):TTexture;  virtual; abstract;
       function loadSound(name:string):TSound;  virtual; abstract;
      public
       function getMeshByName(name:string):TMesh;override;
       function getTextureByName(name:string):TTexture;override;
       function getSoundByName(name:string):TSound;override;
       constructor Create;
     end;

    {** This is the only actual media manager,
        that reads data from files. }
    TFileMediaManager = class(TCachedMediaManager)
      private
       function loadMesh(name:string):TMesh; override;
       function loadTexture(name:string):TTexture;  override;
       function loadSound(name:string):TSound;  override;
      public
       constructor Create;
     end;

implementation

function TCachedMediaManager.getMeshByName(name:string):TMesh;
var mesh:TMesh;
begin
 result:=MeshList.getMeshByName(name);
 if result=nil then
  begin
   mesh:=loadMesh(name);
   if mesh<>nil then MeshList.Add(mesh);
   result:=mesh;
  end;
end;


function TCachedMediaManager.getTextureByName(name:string):TTexture;
var m:TTexture;
begin
 result:=TextureList.getTextureByName(name);
 if result=nil then
  begin
   m:=loadTexture(name);
   if m<>nil then TextureList.Add(m);
   result:=m;
  end;
end;

function TCachedMediaManager.getSoundByName(name:string):TSound;
var m:TSound;
begin
 result:=SoundList.getSoundByName(name);
 if result=nil then
  begin
   m:=loadSound(name);
   if m<>nil then SoundList.Add(m);
   result:=m;
  end;
end;


function TFileMediaManager.loadMesh(name:string):TMesh;
var m:TMesh;
begin
// does this work on exception?

 try
  m:=TGlMesh.Create;
  m.LoadFrom(MEDIA_PATH+name+'.mesh');
  result:=m;
 except
  result:=nil;
 end;
 if result = nil then m.free;
end;


function TFileMediaManager.loadTexture(name:string):TTexture;
var m:TTexture;
begin
 try
  m:=TTexture.Create;
  m.LoadFrom(MEDIA_PATH+name+'.png');
  m.name:=name;
  result:=m;
 except
  result:=nil;
 end;
 if result=nil then m.free;
end;

function TFileMediaManager.loadSound(name:string):TSound;
begin
 result:=TSound.Create(MEDIA_PATH+name+'.wav');
// result.LoadFromFile();
 result.name:=name;
end;


constructor TCachedMediaManager.Create;
begin
	inherited Create;
	MeshList:=TMeshList.Create;
        SoundList:=TSoundList.Create;
	TextureList:=TTextureList.create;
end;

constructor TFileMediaManager.Create;
begin
	inherited Create;
end;

end.
