{** This unit implements some thread utilities. It's based on pthreads on linux.}
unit threadunit;

interface

uses
  Classes, SysUtils, libc ;


type

{** This kind of exception is raised by TMutex methods when something goes wrong }
MutexException = class(Exception)
end;

{** This kind of exception is raised by TMSXThread methods when something goes wrong }
ThreadException = class(Exception)
end;

{** This is a mutex. It can be used for thread syncronization. }
TMutex = class
 private
   m:TPthreadMutex;
 public
   {** Create a new mutex ready for use }
   constructor Create;
   {** Destroy a mutex. Make sure the mutex is not locked before calling this. }
   destructor Destroy;override;
   {** Acquire the mutex, locking it. This call blocks if the mutex is already locked by someone else.
   A Thread can aquire a mutex multiple time. In this case an equal number of Release must be called to
   actually releasing it. }
   procedure Acquire;
   {** Try to acquire the mutex. If it's already locked, it returns false, else it acquire the mutex and return true.
   This call never blocks. }
   function TryAcquire:boolean;
   {** Release a previously acquired mutex }
   procedure Release;
   
end;

{** This is a class that represent a thread. To use it, you must create a subclass that implements the Execute method. }
TMSXThread = class
  private
    fThreadid:pthread_t;
    fFreeOnTerminate:boolean;
    fTerminated:boolean;

    fIsExited:boolean;
  public
    {** This is the main body of the thread. You must override this method to make the thread so something useful }
    procedure Execute; virtual; abstract;
    {** Create a new thread. This only creates the class, it doesn't start the thread yet. }
    constructor Create;
    {** Start the thread. This call return immediately and the new thread is spawned. }
    procedure Run;
    {** Wait for this thread to exit. This can be call by other threads. }
    procedure WaitFor;
    {** Set the terminated flag at true. Thread bodys should check this flag from time to time and
    exit when it's true. There is no way to force the termination of a thread, since it is inherently unsafe:
    resources and locks could be owned by the thread, resulting in leaks, deadlocks and the like. }
    procedure Terminate;
    {** Tells whether the thread has been segnaled to terminate (with the Terminate method) }
    property Terminated:boolean read fTerminated;
    {** Tells is the class should be destroyed when the threads terminate. This is false by default.
    Be carefull using this: you can run in trouble for example with WaitFor or with other syncronization task, since the
    TMSXThread can be 'free'd at any time (from other thread's point of view). }
    property FreeOnTerminate:boolean read fFreeOnTerminate write fFreeOnTerminate;
end;

implementation

function StaticRun(para1:pointer): integer; cdecl;
var t:TMSXThread;
begin
writeln('creating thread');
t:=TMSXThread(para1);
t.Execute;
if t.FreeOnTerminate then t.free else t.fIsExited:=true;
writeln('closing thread');
result:=0;
pthread_exit(nil);
end;

constructor TMSXThread.Create;
begin
fFreeOnTerminate:=false;
fIsExited:=false;
end;

procedure TMSXThread.WaitFor;
var p:pointer;
begin

if fIsExited then exit; // already finished, no need to wait

if pthread_join(fThreadid, p) <> 0 then
   raise ThreadException.create('Error joining thread');
end;

procedure TMSXThread.Run;
var attr:pthread_attr_t;
begin
(* Initialize and set thread detached attribute *)
try
  pthread_attr_init(attr);
  pthread_attr_setdetachstate(attr, PTHREAD_CREATE_JOINABLE);
  if pthread_create(fThreadid, attr,StaticRun, self) <> 0 then
    raise ThreadException.create('Error creating thread');
finally
  pthread_attr_destroy(attr);
end;
end;

procedure TMSXThread.Terminate;
begin
fTerminated:=true;
end;


constructor TMutex.Create;
var
  MAttr : TMutexAttribute;
begin
  try
    if pthread_mutexattr_init(@MAttr)<>0 then raise MutexException.create('Unable to create mutex attrib');
    if pthread_mutexattr_settype(@MAttr, PTHREAD_MUTEX_RECURSIVE)<>0 then raise MutexException.create('Unable to set mutex type');
    if pthread_mutex_init(@m,@MAttr)<>0 then raise MutexException.create('Unable to create mutex');
  finally
      pthread_mutexattr_destroy(@MAttr);
  end;

end;

destructor TMutex.Destroy;
begin
  if pthread_mutex_destroy(@m)<>0 then raise MutexException.create('Unable to destroy mutex');
end;

procedure TMutex.Acquire;
begin
  if pthread_mutex_lock(@m)<>0 then raise MutexException.create('Unable to lock mutex');
end;

function TMutex.TryAcquire:boolean;
var r:longint;
begin
  r:=pthread_mutex_trylock(@m);
  case r of
    0: result:=true;
    EBUSY:result:=false;
    else raise MutexException.create('Unable to lock mutex');
  end;

end;

procedure TMutex.Release;
begin
  if pthread_mutex_unlock(@m)<>0 then raise MutexException.create('Unable to unlock mutex');
end;


end.

