unit gameserverunit;

interface

uses
  Classes, SysUtils, baseunit, commonunit, game, networkunit, messageunit, serverunit, logger;
  
type


 {** An user connected to the server, that autenticated throug the account system.
  It appear on the connected user list on the chat. }
 TUser = class
   {** username loaded from the account }
   username:string;
   {** The user connection to the server }
   connection:TSocket;
   {** a list of the items and units possessed by this player.
      We only need the models becouse we are still off-game,
      in the chat }
   items:TItemStaticInfoList;
   units:TUnitStaticInfoList;
   constructor Create;
   {** Load the current user. If all is OK return empty string, else an error message }
   function Load(user:string; password:string):string;
   destructor Destroy;override;
   procedure Send(const s:string);
 end;

 TUserList = class(TList)
      function getUser(index:integer):TUser;
      property User[index:integer]:TUser read getUser; default;
      function getUserByConnection(conn:TSocket):TUser;
 end;
 
 {** This is the class that keep information of
  created but not yet started game on the server.
  Informations are players, units, items, etc. }
 TGameDescription = class
   {** The game unique identifier }
   id:integer;
   {** The name given by the player }
   name:string;
   {** users taking place to this game }
   Users:TUserList;
   {** the map choosen for this game }
   MapName:string;
   {** Max points to build up the party }
   Points:integer;
   
   Constructor Create;
   destructor destroy; override;
 end;

 TGameDescriptionList = class(TList)
 end;


 {** The main game server program class. }
 TGameServer = class
  private
   fNextGameId:integer;
  public
   {** all connected user }
   ConnectedUsers:TUserList;
   {** all games currently created }
   GameDescriptions:TGameDescriptionList;
   {** running games, with full state }
   Games:TGameList;
   
   PacketServer:TPacketServer;
   
   {** Do jobs on new incoming connection }
   procedure DoConnection(conn:TSocket);
   {** Handles a disconnection }
   procedure DoCloseConnection(conn:TSocket);
   {** This takes all incoming messages from clients, and dispatch them as needed }
   procedure DoMessage(const Message:string; conn:TSocket);
   {** This gets called when a networks event occour, such as new connections, disconnections, data avaiable, etc }
   procedure OnEvent(sender:TPacketServer; conn:TSocket;  EventType:TEventType; text:string);
   
   {** Remove an user from the game server. This handles everithing: notification to other users, 
   managing the game where the user was, etc etc }
   procedure RemoveUser(User:TUser);
   {** Autenticate a user on the server. This handles everything: notifications, user data loading, etc etc }
   procedure AutenticateUser(User,Pass:string; conn:TSocket);
   
   {** This handles messages from logged users in chat section }
   procedure ParseLoggedMessage(const message:string; user:TUser);
   {** This handles messages coming from clients that didn't logon yet }
   procedure ParseUnloggedMessage(const message:string; conn:TSocket);
	{** This handles messages coming from users that are inside a running game }
   procedure ParseInGameMessage(const message:string; game:TGame; user:TUser);
      
   {** Send the whole list of games to a user. Called on login. }
   procedure SendGameList(user:TUser);
   
   {** Creates a new game description, for example when a player require it }
   procedure CreateGameDescription(user:TUser; gamename, mapname, points:string);
   {** Removes a game description from the server. This can happend when a game start or when all users exit from it }
   procedure RemoveGameDescription(gd:TGameDescription);
   
   procedure UserEnterGameDescription(user:TUser; gameid:integer);
   procedure UserExitGameDescription(user:TUser);
   
   {** Send a message to all connected users }
   procedure NotifyAll(const message:string);
   
   {** Get the running game where the user is on, or nil if the user is not in a game }
   function GetGameByUser(user:TUser):TGame;
   {** Get the next game ID, that is used to keep a unique identifier for games }
   function GetNextGameId:integer;
   
   constructor Create;
   
   {** Start the game server }
   procedure Start;
 end;



implementation

procedure TGameServer.UserEnterGameDescription(user:TUser; gameid:integer);
begin
 // notifyAll( msg( M_ENTER_GAME, [user.username, inttostr(gd.id)]));
 // @todo
end;

procedure TGameServer.UserExitGameDescription(user:TUser);
begin
 // @todo 
end;


procedure TGameServer.DoConnection(conn:TSocket);
begin
	 // do nothing special here..
   Log('Connected '+conn.ip);
end;

procedure TGameServer.DoCloseConnection(conn:TSocket);
var user:TUser;
begin
	// find the user disconnecting
	user:=ConnectedUsers.getUserByConnection(conn);
	if user = nil then
	 begin
	 	 // the user was already removed "gracefully" by a tQuit message or other mean.
	 	 // so nothing to do here.
     Log('Gracefully disconnected '+conn.ip);
	 end
  else
   begin
   	 // the user disconnected brutally. this may be due to connection broken
   	 // or malicious disconnection. Handle eventual punition for disconnection here
         Log('Brutally disconnecting '+user.username+' IP: '+conn.ip);
   	 removeUser(user);

   end;
	
end;

procedure TGameServer.RemoveUser(User:TUser);
var i:integer;
    g:TGameDescription;
begin
  log('Removing user '+user.username);
  
  // handle eventual game the user was in.
  for i:=GameDescriptions.count-1 downto 0 do
    begin
      g:=TGameDescription(GameDescriptions[i]);
      if g.Users.IndexOf(user)=0 then
        begin
          // it's the creator of the game, so remove the game
          RemoveGameDescription(g);
        end
      else
        begin
          // it's the opponent, so remove only him
          // @todo
        end;
    end;

  // remove from active users
  ConnectedUsers.remove(user);
  
  // notify remaining people of his departure
  NotifyAll(msg(M_EXIT, [user.Username]));
  
  // close the connection if it's not already closed
  // @todo discover why this blocks!
  // if User.Connection.Active then User.Connection.Close;
  User.free;
  
end;

function TGameServer.GetNextGameId:integer;
begin
	inc(fnextGameId);
	result:=fnextGameId;
end;

procedure TGameServer.RemoveGameDescription(gd:TGameDescription);
begin
  // notify all about game removal
 NotifyAll(msg(M_REMOVE_GAME, [inttostr(gd.id)]));

 GameDescriptions.remove(gd);

 gd.free;
end;

procedure TGameServer.CreateGameDescription(user:TUser; gamename, mapname, points:string);
var gd:TGameDescription;
begin
	gd:=TGameDescription.create;
	gd.mapname:=mapname;
	gd.points:=strtoint(points); // @todo control if it's a number, return error otherwise
	gd.id:=GetNextGameId;
	gd.users.add(user);
	gd.name:=gamename;
	
        //sleep(3000);
 
	// notify all about game creation
	notifyAll(msg(M_CREATE_GAME, [inttostr(gd.id), gd.name, gd.mapName, inttostr(gd.points), gd.users[0].username] ));
	
	GameDescriptions.add(gd);
end;

procedure TGameServer.ParseLoggedMessage(const message:string; user:TUser);
var m:TMsg;
begin
  m:=TMsg.create(message);
	case m.getType of
 	  M_CHAT:
      begin
        Log('#'+user.username+': '+m[1]);
        NotifyAll(msg(M_CHAT, [user.username, m[1]]));
      end;
 	  M_CREATE_GAME:
      begin
        CreateGameDescription(user, m[3], m[1], m[2]);
      end;
 	  M_QUIT:
 	    begin
 	    	RemoveUser(user);
                user:=nil;
 	    end;
 	  else
 	   begin
	   	 // unexpected message..
	   	 // Bug ? Version incompatibility ? Hacking attempt ?
 	   	 // eventually do some log
 	   	 
 	   	 // kick him away
             if user<>nil then RemoveUser(user);
 	   end;
  end;
	m.free;
end;

procedure TGameServer.SendGameList(user:TUser);
var i:integer;
		g:TGameDescription;
begin
	for i:=0 to GameDescriptions.count-1 do
	 begin
	 	g:=TGameDescription(GameDescriptions[i]);
   	notifyAll(msg(M_CREATE_GAME, [inttostr(g.id), g.name, g.mapName, inttostr(g.points), g.users[0].username] ));
	 end;
end;

procedure TGameServer.NotifyAll(const message:string);
var i:integer;
begin
	for i:=0 to ConnectedUsers.count-1 do
	 begin
	 	ConnectedUsers[i].send(message);
	 end;
end;

procedure TGameServer.OnEvent(sender:TPacketServer; conn:TSocket; EventType:TEventType; text:string);
begin
case EventType of
 eMessage:DoMessage(text, conn);
 eConnection:DoConnection(conn);
 eDisconnection:DoCloseConnection(conn);
end;
end;

procedure TGameServer.AutenticateUser(User,Pass:string; conn:TSocket);
var u:TUser;
	  i:integer;
    s:String;
begin
  Log('Autenticating user: '+user+' pass: '+pass);
  u:=TUser.create;
  // link to his connection
  u.connection:=conn;

  s:=u.load(user, pass);
  if s='' then
   begin
    // login ok

    // send confirm for login
    u.send(msg(M_LOGIN_OK, [u.UserName]));

    // notify people of his arrival
    NotifyAll(msg(M_ENTER, [u.UserName]));
    
    // send userlist to the new player.
    for i:=0 to ConnectedUsers.count-1 do
     begin
      u.send(msg(M_USERLIST, [ConnectedUsers[i].UserName]));
     end;
     
    // send current game description list (ie all opened unstarted games ) 
    SendGameList(u); 

    // join the crowd
    ConnectedUsers.add(u);
    Log('User autenticated '+user+' ip: '+conn.ip);
   end	
  else
   begin
   	// no login

    Log('User NOT autenticated '+user+' ip: '+conn.ip);
    // send negative login
    u.send(msg(M_LOGIN_NO, [s]));
    
   	u.free;
   end;
end;

procedure TGameServer.ParseInGameMessage(const message:string; game:TGame; user:TUser);
begin
//
end;

procedure TGameServer.ParseUnloggedMessage(const message:string; conn:TSocket);
var m:TMsg;
begin
	m:=TMsg.create(message);
 
	case m.getType of
 	  M_LOGIN:
 	   begin 
       if m.numParam = 2 then
  	     AutenticateUser(m[1], m[2], conn)
       else
         conn.send(msg(M_LOGIN_NO, ['Enter user and passwd.']));
 	   end;
 	  M_REGISTER: begin end;
 	  else
 	   begin
	   	 // unexpected message, drop the connection
 	   	 // eventually do some log
       Log('Unexpected message "'+message+'"');
 	     conn.Close;
 	   end;
  end;
  
  m.free;

end;

procedure TGameServer.DoMessage(const message:string; conn:TSocket );
var user:TUser;
		game:TGame;
begin
	// find the user sending the message
	user:=ConnectedUsers.getUserByConnection(conn);
	
	if user = nil then
	  begin
	  	// no user found. this is a new connection
	  	ParseUnloggedMessage(message, conn);
	  end
	 else
	  begin
	  	// find if the user is inside a game
	  	game:=getGameByUser(user);
	  	
	  	if game<>nil then 
	  		// user already logged that is inside a running game.
	  		ParseInGameMessage(message, game, user)
	  	else
	  		// user already logged, in the chat section (chat, party operations, game management etc)
	  		ParseLoggedMessage(message, user);
	  end;
	  
end;

constructor TGameServer.Create;
begin
ConnectedUsers:=TUserList.create;
GameDescriptions:=TGameDescriptionList.create;
fNextGameId:=100;
end;

procedure TGameServer.Start;
begin
 PacketServer:=TPacketServer.create(100);
 PacketServer.OnEvent:=OnEvent;
 PacketServer.Start(50401, 1024);
end;

function TUserList.getUser(index:integer):TUser;
begin
	result:=TUser(items[index]);
end;

function TUserList.getUserByConnection(conn:TSocket):TUser;
var i:integeR;
begin
	for i:=0 to count-1 do
	 if user[i].connection = conn then
	   begin
	   	 result:=user[i];
	   	 exit;
	   end;
  result:=nil;	   
end;

constructor TUser.Create;
begin
   items:=TItemStaticInfoList.create;
   units:=TUnitStaticInfoList.create;
end;
destructor TUser.Destroy;
begin
   items.free;
   units.free;
   inherited destroy;
end;

function TUser.Load(user:string; password:string):string;
begin
	// handle user loading and autentication here
	username:=user;
  result:='';
  if user[1] = 'X' then result := 'Invalid user/pass';
end;

procedure TUser.Send(const s:string);
begin
  connection.send(s);
end;

function TGameServer.GetGameByUser(user:TUser):TGame;
begin
	result:=nil;
end;

Constructor TGameDescription.Create;
begin
	Users:=TUserList.create;
end;

destructor TGameDescription.destroy;
begin
	Users.free;
	inherited Destroy;
end;

end.

